import 'package:flutter/material.dart';
import 'package:flutter_gadgets_2/flutter_gadgets_2.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:provider/provider.dart';

import '../utils.dart';

class MockController extends ControllerGadget {
  void onTapWithUpdate() {
    model.updateValue<MockValue>(updateFunction: (v) => v..increment());
  }

  void onTapWithoutUpdate() {
    final value = model.getValue<MockValue>()!;
    value.increment();
  }

  void onTapWithNotify() {
    final value = model.getValue<MockValue>()!;
    value.increment();
    model.notifyFor<MockValue>();
  }

  void setNullableStateToNull() {
    model.updateValue<MockValue>(
        updateFunction: (v) => v..nullableState = null);
  }
}

void main() {
  testWidgets('SubscriberWidget value', (tester) async {
    await tester.pumpWidget(AppGadget(
      init: (model) => model.putValue<MockValue>(value: MockValue()),
      child: SubscriberGadget<MockValue>(
        builder: (_, value) {
          return Text(
            '${value.state}',
            textDirection: TextDirection.ltr,
          );
        },
      ),
    ));
    expect(find.text('0'), findsOneWidget);

    await tester.pumpWidget(Container());
  });

  testWidgets('SubscriberWidget value changed with update', (tester) async {
    await tester.pumpWidget(AppGadget(
      init: (model) => model.putValue<MockValue>(value: MockValue()),
      child: ViewGadget<MockController>(
        controller: MockController(),
        builder: (_, controller) => SubscriberGadget<MockValue>(
          builder: (_, value) {
            return GestureDetector(
              onTap: controller.onTapWithUpdate,
              child: Text(
                '${value.state}',
                textDirection: TextDirection.ltr,
              ),
            );
          },
        ),
      ),
    ));

    await tester.tap(find.text('0'));
    await tester.pump();

    final observableModel = Provider.of<ObservableModel>(
      tester.element(find.byType(Text)),
      listen: false,
    );

    expect(
      observableModel.getValue<MockValue>()!.state,
      1,
    );
    expect(find.text('1'), findsOneWidget);

    await tester.pumpWidget(Container());
  });

  testWidgets('SubscriberWidget value changed without update', (tester) async {
    final key = UniqueKey();
    await tester.pumpWidget(AppGadget(
      init: (model) => model.putValue<MockValue>(value: MockValue()),
      child: ViewGadget<MockController>(
        controller: MockController(),
        builder: (_, controller) => SubscriberGadget<MockValue>(
          builder: (_, bean) {
            return GestureDetector(
              onTap: controller.onTapWithoutUpdate,
              child: Text(
                '${bean.state}',
                key: key,
                textDirection: TextDirection.ltr,
              ),
            );
          },
        ),
      ),
    ));

    final model = Provider.of<ObservableModel>(
      tester.element(find.byType(Text)),
      listen: false,
    );

    final bean = model.getValue<MockValue>()!;
    bean.increment();
    expect(model.getValue<MockValue>()!.state, 1);

    await tester.tap(find.byKey(key));
    await tester.pump();
    expect(find.text('1'), findsNothing);

    await tester.pumpWidget(Container());
  });

  testWidgets('SubscriberWidget value changed with notify', (tester) async {
    await tester.pumpWidget(AppGadget(
      init: (model) => model.putValue<MockValue>(value: MockValue()),
      child: ViewGadget<MockController>(
        controller: MockController(),
        builder: (_, controller) => SubscriberGadget<MockValue>(
          builder: (_, value) {
            return GestureDetector(
              onTap: controller.onTapWithNotify,
              child: Text(
                '${value.state}',
                textDirection: TextDirection.ltr,
              ),
            );
          },
        ),
      ),
    ));

    await tester.tap(find.text('0'));
    await tester.pump();

    final observableModel = Provider.of<ObservableModel>(
      tester.element(find.byType(Text)),
      listen: false,
    );

    expect(
      observableModel.getValue<MockValue>()!.state,
      1,
    );
    expect(find.text('1'), findsOneWidget);

    await tester.pumpWidget(Container());
  });

  testWidgets('SubscriberWidget properties selector', (tester) async {
    await tester.pumpWidget(AppGadget(
      init: (model) => model.putValue<MockValue>(value: MockValue()),
      child: ViewGadget<MockController>(
        controller: MockController(),
        builder: (_, controller) => SubscriberGadget<MockValue>(
          propertiesSelector: (value) => [value.nullableState],
          builder: (_, value) {
            return GestureDetector(
              onTap: controller.setNullableStateToNull,
              child: Text(
                value.nullableState?.toString() ?? '-',
                textDirection: TextDirection.ltr,
              ),
            );
          },
        ),
      ),
    ));

    await tester.tap(find.text('100'));
    await tester.pump();

    final observableModel = Provider.of<ObservableModel>(
      tester.element(find.byType(Text)),
      listen: false,
    );

    expect(
      observableModel.getValue<MockValue>()!.nullableState,
      isNull,
    );
    expect(find.text('-'), findsOneWidget);

    await tester.pumpWidget(Container());
  });
}
